from .utilities import numberToHex, hexToByteArray
from .signer import Signer
from .interfaces import Order
import hashlib


class OrderSigner(Signer):
    def __init__(self, version="1.0"):
        super().__init__()
        self.version = version

    def get_order_flags(self, order):
        """0th bit = ioc
        1st bit = postOnly
        2nd bit = reduceOnly
        3rd bit  = isBuy
        4th bit = orderbookOnly
        e.g. 00000000 // all flags false
        e.g. 00000001 // ioc order, sell side, can be executed by taker
        e.e. 00010001 // same as above but can only be executed by settlement operator
        """
        flag = 0
        if order["ioc"]:
            flag += 1
        if order["postOnly"]:
            flag += 2
        if order["reduceOnly"]:
            flag += 4
        if order["isBuy"]:
            flag += 8
        if order["orderbookOnly"]:
            flag += 16
        return flag

    def get_order_hash(self, order: Order, withBufferHex=True):
        """
        Returns order hash.
        Inputs:
            - order: the order to be signed
        Returns:
            - str: order hash
        """
        flags = self.get_order_flags(order)
        flags = hexToByteArray(numberToHex(flags, 2))

        buffer = bytearray()
        orderPriceHex = hexToByteArray(numberToHex(int(order["price"])))
        orderQuantityHex = hexToByteArray(numberToHex(int(order["quantity"])))
        orderLeverageHex = hexToByteArray(numberToHex(int(order["leverage"])))
        orderSalt = hexToByteArray(numberToHex(int(order["salt"])))
        orderExpiration = hexToByteArray(numberToHex(int(order["expiration"]), 16))
        orderMaker = hexToByteArray(numberToHex(int(order["maker"], 16), 64))
        orderMarket = hexToByteArray(numberToHex(int(order["market"], 16), 64))
        bluefin = bytearray("Bluefin", encoding="utf-8")

        buffer = (
            orderPriceHex
            + orderQuantityHex
            + orderLeverageHex
            + orderSalt
            + orderExpiration
            + orderMaker
            + orderMarket
            + flags
            + bluefin
        )

        # for cancel order signature verification we use buffer directly
        # for placing order we use buffer.hex().encode("utf-8")
        if withBufferHex:
            order_hash = hashlib.sha256(buffer.hex().encode("utf-8")).digest()
        else:
            order_hash = hashlib.sha256(buffer).digest()
        return order_hash

    def sign_order(self, order: Order, private_key):
        """
        Used to create an order signature. The method will use the provided key
        in params to sign the order.

        Args:
            order (Order): an order containing order fields (look at Order interface)
            private_key (str): private key of the account to be used for signing

        Returns:
            str: generated signature
        """
        order_hash = self.get_order_hash(order)
        return self.sign_hash(order_hash, private_key, "")



import requests
import json


def rpc_unsafe_moveCall(
    url,
    params,
    function_name: str,
    function_library: str,
    userAddress,
    packageId,
    gasBudget=100000000,
):
    """
    Does the RPC call to SUI chain
    Input:
      url: url of the node
      params: a list of arguments to be passed to function
      function_name: name of the function to call on sui
      function_library: name of the library/module in which the function exists
      userAddress: Address of the signer
      packageId: package id in which the module exists
      gasBudget(optional): gasBudget defaults to 100000000

    Output:
      Returns the request form serialised in bytes ready to be signed.

    """
    base_dict = {}
    base_dict["jsonrpc"] = "2.0"
    base_dict["id"] = 1689764924887
    base_dict["method"] = "unsafe_moveCall"
    base_dict["params"] = []
    base_dict["params"].extend(
        [userAddress, packageId, function_library, function_name]
    )
    base_dict["params"].append([])
    base_dict["params"].append(params)

    base_dict["params"].append(None)
    base_dict["params"].append(str(gasBudget))

    payload = json.dumps(base_dict)

    headers = {"Content-Type": "application/json"}
    response = requests.request("POST", url, headers=headers, data=payload)
    result = json.loads(response.text)
    return result["result"]["txBytes"]


def rpc_sui_executeTransactionBlock(url, txBytes, signature):
    """
    Execute the SUI call on sui chain
    Input:
      url: url of the node
      txBytes: the call in serialised form
      signature: txBytes signed by signer

    Output:
      result of transaction


    """
    base_dict = {}
    base_dict["jsonrpc"] = "2.0"
    base_dict["id"] = 5
    base_dict["method"] = "sui_executeTransactionBlock"
    base_dict["params"] = []
    base_dict["params"].append(txBytes)
    base_dict["params"].append([signature])

    outputTypeDict = {
        "showInput": True,
        "showEffects": True,
        "showEvents": True,
        "showObjectChanges": True,
    }
    base_dict["params"].append(outputTypeDict)
    base_dict["params"].append("WaitForLocalExecution")
    payload = json.dumps(base_dict)

    headers = {"Content-Type": "application/json"}
    response = requests.request("POST", url, headers=headers, data=payload)
    result = json.loads(response.text)
    return result


def rpc_call_sui_function(url, params, method="suix_getCoins"):
    """
    for calling sui functions:
    Input:
      url: url of node
      params: arguments of function
      method(optional): the name of method in sui we want to call. defaults to suix_getCoins
    """
    base_dict = {}
    base_dict["jsonrpc"] = "2.0"
    base_dict["id"] = 1
    base_dict["method"] = method
    base_dict["params"] = params
    payload = json.dumps(base_dict)

    headers = {"Content-Type": "application/json"}
    response = requests.request("POST", url, headers=headers, data=payload)
    result = json.loads(response.text)
    return result["result"]



from .enumerations import MARKET_SYMBOLS


_DEFAULT_MARKET = "BTC-PERP"


class Contracts:
    def __init__(self):
        self.contracts_global_info = {}
        self.contract_info = {}

    def set_contract_addresses(self, contracts_info):
        self.contract_info = contracts_info
        self.contracts_global_info = contracts_info["auxiliaryContractsAddresses"][
            "objects"
        ]

    def get_sub_account_id(self):
        return self.contracts_global_info["SubAccounts"]["id"]

    def get_bank_table_id(self):
        return self.contracts_global_info["BankTable"]["id"]

    def get_package_id(self):
        return self.contracts_global_info["package"]["id"]

    def get_bank_id(self):
        return self.contracts_global_info["Bank"]["id"]

    def get_currency_type(self):
        return self.contracts_global_info["Currency"]["dataType"]

    def get_price_oracle_object_id(self, market: MARKET_SYMBOLS):
        return self.contract_info[market.value]["PriceOracle"]["id"]

    def get_perpetual_id(self, market: MARKET_SYMBOLS):
        return self.contract_info[market.value]["Perpetual"]["id"]



from typing import TypedDict, List
from .enumerations import *


class Order(TypedDict):
    market: str
    price: int
    isBuy: bool
    reduceOnly: bool
    quantity: int
    postOnly: bool
    orderbookOnly: bool
    leverage: int
    expiration: int
    salt: int
    maker: str
    ioc: bool


class SignedOrder(Order):
    typedSignature: str


class RequiredOrderFields(TypedDict):
    # symbol: MARKET_SYMBOLS # market for which to create order
    market: str
    price: int  # price at which to place order. Will be zero for a market order
    quantity: int  # quantity/size of order
    side: ORDER_SIDE  # BUY/SELL
    orderType: ORDER_TYPE  # MARKET/LIMIT


class OrderSignatureRequest(RequiredOrderFields):
    leverage: int  # (optional) leverage to take, default is 1
    reduceOnly: bool  # (optional)  is order to be reduce only true/false, default its false
    salt: int  # (optional)  random number for uniqueness of order. Generated randomly if not provided
    expiration: int  # (optional) time at which order will expire. Will be set to 1 month if not provided
    maker: str  # (optional) maker of the order, if not provided the account used to initialize the client will be default maker
    # isBuy: bool
    postOnly: bool
    orderBookOnly: bool
    ioc: bool


class OrderSignatureResponse(RequiredOrderFields):
    maker: str
    orderSignature: str


class PlaceOrderRequest(OrderSignatureResponse):
    timeInForce: TIME_IN_FORCE  # FOK/IOC/GTT by default all orders are GTT
    postOnly: bool  # true/false, default is true
    cancelOnRevert: bool  # if true, the order will be cancelled in case of on-chain settlement error, default is false
    clientId: str  # id of the client


class GetOrderbookRequest(TypedDict):
    symbol: MARKET_SYMBOLS
    limit: int  # number of bids/asks to retrieve, should be <= 50


class OnboardingMessage(TypedDict):
    action: str
    onlySignOn: str


class OrderResponse(TypedDict):
    id: int
    clientId: str
    requestTime: int
    cancelReason: CANCEL_REASON
    orderStatus: ORDER_STATUS
    hash: str
    symbol: MARKET_SYMBOLS
    orderType: ORDER_TYPE
    timeInForce: TIME_IN_FORCE
    userAddress: str
    side: ORDER_SIDE
    price: str
    quantity: str
    leverage: str
    reduceOnly: bool
    expiration: int
    salt: int
    orderSignature: str
    filledQty: str
    avgFillPrice: str
    createdAt: int
    updatedAt: int
    makerFee: str
    takerFee: str
    openQty: str
    cancelOnRevert: bool


class GetOrderResponse(OrderResponse):
    fee: str
    postOnly: bool
    triggerPrice: str


class GetCandleStickRequest(TypedDict):
    symbol: MARKET_SYMBOLS
    interval: Interval
    startTime: float
    endTime: float
    limit: int


class GetMarketRecentTradesRequest(TypedDict):
    symbol: MARKET_SYMBOLS
    pageSize: int
    pageNumber: int
    traders: TRADE_TYPE


class OrderCancelSignatureRequest(TypedDict):
    symbol: MARKET_SYMBOLS
    hashes: list
    parentAddress: str  # (optional) should only be provided by a sub account


class OrderCancellationRequest(OrderCancelSignatureRequest):
    signature: str


class CancelOrder(TypedDict):
    hash: str
    reason: str


class CancelOrderResponse(TypedDict):
    message: str
    data: dict


class GetTransactionHistoryRequest(TypedDict):
    symbol: MARKET_SYMBOLS  # will fetch orders of provided market
    pageSize: int  # will get only provided number of orders must be <= 50
    pageNumber: int  # will fetch particular page records. A single page contains 50 records.


class GetPositionRequest(GetTransactionHistoryRequest):
    parentAddress: str  # (optional) should be provided by sub accounts


class GetUserTradesRequest(TypedDict):
    symbol: MARKET_SYMBOLS
    maker: bool
    fromId: int
    startTime: int
    endTime: int
    pageSize: int
    pageNumber: int
    type: ORDER_TYPE
    parentAddress: str  # (optional) should be provided by sub account


class GetOrderRequest(GetTransactionHistoryRequest):
    statuses: List[ORDER_STATUS]  # (optional) status of orders to be fetched
    orderId: int  # (optional) the id of order to be fetched
    orderType: List[ORDER_TYPE]  # (optional) type of order Limit/Market
    orderHashes: List[str]  # (optional) hashes of order to be fetched
    parentAddress: str  # (optional) should be provided by sub accounts


class GetFundingHistoryRequest(TypedDict):
    symbol: MARKET_SYMBOLS  # will fetch orders of provided market
    pageSize: int  # will get only provided number of orders must be <= 50
    cursor: int  # will fetch particular page records. A single page contains 50 records.
    parentAddress: str  # (optional) should be provided by a sub account


class FundingHistoryResponse(TypedDict):
    id: int  # unique id
    symbol: MARKET_SYMBOLS  # market for which to create order
    userAddress: str  # user public address
    quantity: int  # size of position
    time: int  # created time
    appliedFundingRate: str  # funding rate percent applied
    isFundingRatePositive: bool  # was funding rate +ve or -ve
    payment: str  # amount
    isPaymentPositive: bool  # whether payment was deducted or added
    oraclePrice: str  # price from oracle
    side: ORDER_SIDE  # BUY/SELL
    blockNumber: int  # transaction block number
    isPositionPositive: bool  # is position LONG or SHORT


class GetFundingHistoryResponse(TypedDict):
    isMoreDataAvailable: bool  # boolean indicating if there is more data available
    nextCursor: int  # next page number
    data: List[FundingHistoryResponse]


class GetTransferHistoryRequest(TypedDict):
    pageSize: int  # will get only provided number of orders must be <= 50
    cursor: int  # will fetch particular page records. A single page contains 50 records.
    action: str  # (optional) Deposit / Withdraw


class UserTransferHistoryResponse(TypedDict):
    id: int  # unique id
    status: str  # status of transaction
    action: str  # Deposit / Withdraw
    amount: str  # amount withdrawn/deposited
    userAddress: str  # user public address
    blockNumber: int  # transaction block number
    latestTxHash: str  # transaction hash
    time: int  # created time
    createdAt: int
    updatedAt: int


class GetUserTransferHistoryResponse(TypedDict):
    isMoreDataAvailable: bool  # boolean indicating if there is more data available
    nextCursor: int  # next page number
    data: List[UserTransferHistoryResponse]


class CountDown(TypedDict):
    symbol: str
    countDown: int


class GetCancelOnDisconnectTimerRequest(TypedDict):
    symbol: MARKET_SYMBOLS  # will fetch Cancel On Disconnect Timer of provided market
    parentAddress: str  # (optional) should be provided by a sub account


class PostTimerAttributes(TypedDict):
    countDowns: List[CountDown]
    parentAddress: str


class FailedCountDownResetResponse(TypedDict):
    symbol: str
    reason: str


class PostTimerResponse(TypedDict):
    acceptedToReset: List[str]
    failedReset: List[FailedCountDownResetResponse]



Networks = {
    "SUI_STAGING": {
        "url": "https://fullnode.testnet.sui.io:443",
        "apiGateway": "https://dapi.api.sui-staging.bluefin.io",
        "socketURL": "wss://dapi.api.sui-staging.bluefin.io",
        "dmsURL": "https://api.sui-staging.bluefin.io/dead-man-switch",
        "webSocketURL": "wss://notifications.api.sui-staging.bluefin.io",
        "onboardingUrl": "https://testnet.bluefin.io",
    },
    "SUI_PROD": {
        "url": "https://fullnode.testnet.sui.io:443",
        "apiGateway": "https://dapi.api.sui-prod.bluefin.io",
        "socketURL": "wss://dapi.api.sui-prod.bluefin.io",
        "dmsURL": "https://api.sui-prod.bluefin.io/dead-man-switch",
        "webSocketURL": "wss://notifications.api.sui-prod.bluefin.io",
        "onboardingUrl": "https://trade-sui.bluefin.exchange",
    },
}

ORDER_FLAGS = {"IS_BUY": 1, "IS_DECREASE_ONLY": 2}

TIME = {
    "SECONDS_IN_A_MINUTE": 60,
    "SECONDS_IN_A_DAY": 86400,
    "SECONDS_IN_A_MONTH": 2592000,
}

ADDRESSES = {
    "ZERO": "0x0000000000000000000000000000000000000000",
}

SERVICE_URLS = {
    "MARKET": {
        "ORDER_BOOK": "/orderbook",
        "RECENT_TRADE": "/recentTrades",
        "CANDLE_STICK_DATA": "/candlestickData",
        "EXCHANGE_INFO": "/exchangeInfo",
        "MARKET_DATA": "/marketData",
        "META": "/meta",
        "STATUS": "/status",
        "SYMBOLS": "/marketData/symbols",
        "CONTRACT_ADDRESSES": "/marketData/contractAddresses",
        "TICKER": "/ticker",
        "MASTER_INFO": "/masterInfo",
        "FUNDING_RATE": "/fundingRate",
    },
    "USER": {
        "USER_POSITIONS": "/userPosition",
        "USER_TRADES": "/userTrades",
        "ORDERS": "/orders",
        "GENERATE_READONLY_TOKEN": "/generateReadOnlyToken",
        "ACCOUNT": "/account",
        "USER_TRANSACTION_HISTORY": "/userTransactionHistory",
        "AUTHORIZE": "/authorize",
        "ADJUST_LEVERAGE": "/account/adjustLeverage",
        "FUND_GAS": "/account/fundGas",
        "TRANSFER_HISTORY": "/userTransferHistory",
        "FUNDING_HISTORY": "/userFundingHistory",
        "CANCEL_ON_DISCONNECT": "/dms-countdown",
    },
    "ORDERS": {
        "ORDERS": "/orders",
        "ORDERS_HASH": "/orders/hash",
    },
}

SUI_BASE_NUM = 1000000000
DAPI_BASE_NUM = 1000000000000000000
CONTRACTS_BASE_NUM = 1000000



from datetime import datetime
from random import randint

# from web3 import Web3
import time
import bip_utils
import hashlib
from typing import Union
from .constants import SUI_BASE_NUM, DAPI_BASE_NUM, CONTRACTS_BASE_NUM


def toDapiBase(number: Union[int, float]) -> int:
    return int(number * DAPI_BASE_NUM)


def fromDapiBase(number: Union[int, float], dtype=int) -> int:
    return dtype(number / DAPI_BASE_NUM)


def toSuiBase(number: Union[int, float], base=SUI_BASE_NUM) -> int:
    return int(number * base)


def fromSuiBase(number: Union[str, int], base=SUI_BASE_NUM) -> float:
    number = float(number)
    return number / float(base)


def numberToHex(num, pad=32):
    # converting number to Hexadecimal format
    hexNum = hex(num)

    # padding it with zero to make the size 32 bytes
    padHex = hexNum[2:].zfill(pad)
    return padHex


def hexToByteArray(hexStr):
    return bytearray.fromhex(hexStr)


def mnemonicToPrivateKey(seedPhrase: str) -> str:
    bip39_seed = bip_utils.Bip39SeedGenerator(seedPhrase).Generate()
    bip32_ctx = bip_utils.Bip32Slip10Ed25519.FromSeed(bip39_seed)
    derivation_path = "m/44'/784'/0'/0'/0'"
    bip32_der_ctx = bip32_ctx.DerivePath(derivation_path)
    private_key: str = bip32_der_ctx.PrivateKey().Raw()
    return private_key


def privateKeyToPublicKey(privateKey: str) -> str:
    privateKeyBytes = bytes(privateKey)
    bip32_ctx = bip_utils.Bip32Slip10Ed25519.FromPrivateKey(privateKeyBytes)
    public_key: str = bip32_ctx.PublicKey().RawCompressed()
    return public_key


def getAddressFromPublicKey(publicKey: str) -> str:
    address: str = (
        "0x" + hashlib.blake2b(publicKey.ToBytes(), digest_size=32).digest().hex()[:]
    )
    return address


def strip_hex_prefix(input):
    if input[0:2] == "0x":
        return input[2:]
    else:
        return input


def address_to_bytes32(addr):
    return "0x000000000000000000000000" + strip_hex_prefix(addr)


def bn_to_bytes8(value: int):
    return str("0x" + "0" * 16 + hex(value)[2:]).encode("utf-8")


def default_value(dict, key, default_value):
    if key in dict:
        return dict[key]
    else:
        return default_value


def default_enum_value(dict, key, default_value):
    if key in dict:
        return dict[key].value
    else:
        return default_value.value


def current_unix_timestamp():
    return int(datetime.now().timestamp())


def random_number(max_range):
    return current_unix_timestamp() + randint(0, max_range) + randint(0, max_range)


def extract_query(value: dict):
    query = ""
    for i, j in value.items():
        query += "&{}={}".format(i, j)
    return query[1:]


def extract_enums(params: dict, enums: list):
    for i in enums:
        if i in params.keys():
            if type(params[i]) == list:
                params[i] = [x.value for x in params[i]]
            else:
                params[i] = params[i].value
    return params


def config_logging(logging, logging_level, log_file: str = None):
    """Configures logging to provide a more detailed log format, which includes date time in UTC
    Example: 2021-11-02 19:42:04.849 UTC <logging_level> <log_name>: <log_message>
    Args:
        logging: python logging
        logging_level (int/str): For logging to include all messages with log levels >= logging_level. Ex: 10 or "DEBUG"
                                 logging level should be based on https://docs.python.org/3/library/logging.html#logging-levels
    Keyword Args:
        log_file (str, optional): The filename to pass the logging to a file, instead of using console. Default filemode: "a"
    """

    logging.Formatter.converter = time.gmtime  # date time in GMT/UTC
    logging.basicConfig(
        level=logging_level,
        filename=log_file,
        format="%(asctime)s.%(msecs)03d UTC %(levelname)s %(name)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )



from .utilities import *
import base64


class SuiWallet:
    def __init__(self, seed="", privateKey=""):
        if seed == "" and privateKey == "":
            return "Error"
        if seed != "":
            self.privateKey = mnemonicToPrivateKey(seed)
            self.publicKey = privateKeyToPublicKey(self.privateKey)
            self.key = self.getPrivateKey()
        elif privateKey != "":
            self.privateKey = privateKey
            self.publicKey = privateKeyToPublicKey(self.privateKey)
            self.key = self.getPrivateKey()

        else:
            return "error"

        self.publicKeyBase64 = base64.b64encode(self.publicKey.ToBytes()[1:])
        self.privateKeyBase64 = base64.b64encode(self.privateKey.ToBytes()[1:])

        self.privateKeyBytes = self.privateKey.ToBytes()
        self.publicKeyBytes = self.publicKey.ToBytes()

        self.address = getAddressFromPublicKey(self.publicKey)

    def getPublicKey(self):
        return self.publicKey.ToHex()

    def getPrivateKey(self):
        return self.privateKey.ToHex()

    def getUserAddress(self):
        return self.address
